/*******************************************************************************
*    I²C Slave Library 1.0.                                                    *
*    Copyright (C) 2010-averyfarwaydate Luca Bertoncello                       *
*    Hartigstrasse, 12 - 01127 Dresden Deutschland                             *
*    E-Mail: lucabert@lucabert.de, lucabert@lucabert.com                       *
*    http://www.lucabert.de/  http://www.lucabert.com/                         *
*                                                                              *
*    This program is free software; you can redistribute it and/or modify      *
*    it under the terms of the GNU General Public License as published by      *
*    the Free Software Foundation; version 2 of the License.                   *
*                                                                              *
*    This program is distributed in the hope that it will be useful,           *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
*    GNU General Public License for more details.                              *
*                                                                              *
*    You should have received a copy of the GNU General Public License         *
*    along with this program; if not, write to the Free Software               *
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
*******************************************************************************/

#include <util/twi.h>
#include <avr/interrupt.h>
#include <stdint.h>
#include <string.h>
#include "inc/tools.h"
#include "inc/i2cslave.h"

volatile uint8_t nSentData;
volatile uint8_t operation;
volatile uint8_t gotSignal;

// ACK nach empfangenen Daten senden / ACK nach gesendeten Daten erwarten
#define TWCR_ACK    TWCR = (1 << TWEN) | (1 << TWIE) | (1 << TWINT) | (1 << TWEA) | (0 << TWSTA) | (0 << TWSTO) | (0 << TWWC)
// NACK nach empfangenen Daten senden / NACK nach gesendeten Daten erwarten
#define TWCR_NACK   TWCR = (1 << TWEN) | (1 << TWIE) | (1 << TWINT) | (0 << TWEA) | (0 << TWSTA) | (0 << TWSTO) | (0 << TWWC)

/**
 * Initialisierung des I²C-Slave-Subsystems
 *
 * @param uint8_t addr      I²C-Adresse des Gerätes
 */
void initI2Cslave(uint8_t adrr)
{
  registerAddress = NOREGISTERCHOOSED;
  dataCompleted = 0;
  nSentData = 0;
  operation = DUNNO;
  gotSignal = SIG_DUNNO;

// Adresse des Gerätes wird gesetzt
  TWAR = adrr;
// Wir sind SLAVE, also keine Frequenz und Stop bestimmen
  TWCR &= ~(1 << TWSTA) | (1 << TWSTO);
// Acknowledge Bit + TWI aktivieren. Interrupt auflösen bei I²C-Signalen
  TWCR |= (1 << TWEA) | (1 << TWEN) | (1 << TWIE);
// Interrupt aktivieren...
  sei();
}

/**
 * Interruptfunktion, die bei I²C-Signale aufgerufen wird
 */
ISR(TWI_vect)
{
  int l, k;

  switch(TW_STATUS)
  {
// Slave Receiver (START)
// Hier wird das Programm landen, wenn Daten vom Master empfangen werden.
// Es können Daten sind, oder (wenn REGISTERCOUNT > 0) die Adresse des Registers (in dem Fall,
// immer der erste Byte!)
    case TW_SR_SLA_ACK:
      if(gotSignal == SIG_DUNNO ||
         gotSignal == SIG_STOP)
      {
        gotSignal = SIG_START;
        registerAddress = NOREGISTERCHOOSED;
        for(l = 0; l < REGISTERCOUNT + 1; l++)
          receivedDataCount[l] = 0;
        for(l = 0; l < REGISTERCOUNT + 1; l++)
          for(k = 0; k < MAXRECEIVEDDATA; k++)
            receivedData[l][k] = 0;
      }
      dataCompleted = 0;
      TWCR_ACK;
    break;
// Slave Receiver, Daten empfangen
// Je nachdem, ob Register eingerichtet worden sind, werden Daten oder Registeradressen empfangen
    case TW_SR_DATA_ACK:
#if REGISTERCOUNT == 0
// Es gibt keine Register, also werde nur Daten empfangen!
      operation = RECEIVE;
      receivedData[0x00][receivedDataCount[0x00]++] = TWDR;
      if(receivedDataCount[0x00] < (MAXRECEIVEDDATA - 1))
        TWCR_ACK;
      else
        TWCR_NACK;
#else
      if(gotSignal == SIG_START)
      { // Der REPEATED START wurde noch nicht empfangen. Das ist dann nur die Adresse des Registers
        registerAddress = TWDR;
        if(registerAddress > REGISTERCOUNT)
          TWCR_NACK;
        else
          TWCR_ACK;
        operation = DUNNO;
      }
      else
      { // Hier wurde schon die Registernummer empfangen. Jetzt werden nur die Daten für diesen Register empfangen
        operation = RECEIVE;
        receivedData[registerAddress][receivedDataCount[registerAddress]++] = TWDR;
        if(receivedDataCount[registerAddress] < (MAXRECEIVEDDATA - 1))
          TWCR_ACK;
        else
          TWCR_NACK;
      }
#endif
    break;
// Aufgerufen bei STOP oder REPEATED START
    case TW_SR_STOP:
#if REGISTERCOUNT == 0
      if(operation == RECEIVE)
        dataCompleted = 1; // Nun können die Daten vom Programm benutzt werden
      TWCR_ACK;
#else
// Es wurde ein STOP oder ein REPEATED START empfangen
      if(gotSignal == SIG_START &&
         registerAddress != NOREGISTERCHOOSED)
      { // Es wurde ein REPEATED START empfangen
        gotSignal = SIG_REPSTART;
      }
      else
      { // Es wurde ein STOP empfangen
        gotSignal = SIG_STOP;
        if(operation == RECEIVE)
          dataCompleted = 1; // Nun können die Daten vom Programm benutzt werden
        operation = DUNNO;
      }
      TWCR_ACK;
#endif
    break;
// Slave Transmitter (START oder REPEATED START)
// Hier wird das Programm landen, wenn Daten vom Master verlangt werden.
    case TW_ST_SLA_ACK:
      gotSignal = SIG_DUNNO;
      nSentData = 0;
      operation = TRANSMIT;
#if REGISTERCOUNT == 0
// Es gibt keine Register, also es kann nicht passieren, daß das Programm
// hier nach einem REPEATED START landet.
// Es werden die Daten vom REGISTER = 0 geschickt
      registerAddress = 0x00;
#endif
// Und hier werden tatsächlich die Daten versendet
    case TW_ST_DATA_ACK:
      TWDR = toSendData[registerAddress][nSentData++];
      if(nSentData < maxSendData[registerAddress])
        TWCR_ACK;
      else
        TWCR_NACK;
    break;
    default:
      TWCR_ACK;
    break;
  }
}
